# E-mail utilities
#
# A lot is adapted from https://mg.pov.lt/blog/unicode-emails-in-python


# Python Imports
import datetime
import email.charset
import email.header
import email.mime.multipart
import email.mime.text
import email.utils
import json
import logging
import smtplib

# Extern Imports
from tornado.options import options
import aegis.model
import aegis.stdlib
import pytz
import tornado.template

# Project Imports
import config


# Use Quoted-printable encoding instead of base64 since it's far more legible at a glance! From https://bugs.python.org/issue12552
email.charset.add_charset('utf-8', email.charset.SHORTEST, email.charset.QP)

class Mailer(object):
    template_registry = {}


    @classmethod
    def register_template(cls, domain, email_type_name, function):
        cls.template_registry.setdefault(domain, {})
        cls.template_registry[domain][email_type_name] = function


    @classmethod
    def mime(cls, txt, subtype):
        if txt is None:
            return None
        return email.mime.text.MIMEText(txt, subtype, 'UTF-8')


    @classmethod
    def encode_email(cls, sender, recipient, subject, body, reply_to, service_sender):
        plain, html = None, None
        if isinstance(body, str):
            plain = body
        else:
            plain = body.get('plain', None)
            html = body.get('html', None)
        header_charset = 'utf-8'
        parts = []
        if plain:
            parts.append(cls.mime(plain, 'plain'))
        if html:
            parts.append(cls.mime(html, 'html'))
        if len(parts) > 1:
            msg = email.mime.multipart.MIMEMultipart('alternative')
            for p in parts:
                msg.attach(p)
        else:
            msg = parts[0]
        # Set up headers
        headers = (('To', recipient), ('From', sender), ('Reply-to', reply_to),
                   ('Sender', service_sender))
        for header in headers:
            if not header[1]:
                continue
            name, addr = email.utils.parseaddr(header[1])
            name = str(email.header.Header(name, header_charset))
            if header[0] == 'To' and name:
                msg[header[0]] = '"%s" <%s>' % (name.replace('"', "'"), addr)
            else:
                msg[header[0]] = email.utils.formataddr((name, addr))
        # Finish message
        msg['Subject'] = email.header.Header(subject, header_charset)
        return msg.as_string()


    @classmethod
    def render_email(cls, handler, from_email, to_addrs, subject, template, email_opts, **kwargs):
        if not handler:
            host_config = config.hostnames[kwargs['domain']]
            template_loader = tornado.template.Loader(host_config['template_path'])
            plain = ''
            html = ''
            try:
                plain = template_loader.load(template+'.txt').generate(**email_opts)
            except FileNotFoundError as ex:
                if kwargs.get('debug'):
                    logging.exception("Couldn't render plaintext email for template: %s" % template)
            try:
                html = template_loader.load(template+'.html').generate(**email_opts)
            except FileNotFoundError as ex:
                if kwargs.get('debug'):
                    logging.exception("Couldn't render HTML email for template: %s" % template)
        else:
            plain = handler.render_string('%s.txt' % template, **email_opts)
            html =  handler.render_string('%s.html' % template, **email_opts)
        body = {'plain': plain, 'html': html}
        reply_to = kwargs.get('reply_to', from_email)
        service_sender = kwargs.get('service_sender', from_email)
        return cls.encode_email(from_email, to_addrs, subject, body, reply_to, service_sender)


    @classmethod
    def send_mailer(cls, email_tracking_id, dbconn):
        #aegis.stdlib.logw("in mail.send_mailer")
        email_tracking = aegis.model.EmailTracking.get_id(email_tracking_id, dbconn=dbconn)
        email_type = aegis.model.EmailType.get_id(email_tracking['email_type_id'], dbconn=dbconn)
        email_data = json.loads(email_tracking['email_data'])
        # Check to/from emails and format accordingly
        from_email = aegis.model.Email.get_id(email_tracking['from_email_id'], dbconn=dbconn)
        email_data['from_addr'] = from_email['email']
        to_email = aegis.model.Email.get_id(email_tracking['to_email_id'], dbconn=dbconn)
        if to_email['delete_dttm']:
            logging.error("Not sending to deleted email_id: %s", to_email['to_email_id'])
            email_tracking.mark_deleted(dbconn=dbconn)
            return False
        email_data['to_addr'] = to_email['email']
        # Turn this into a (first_name or email) if it's a member
        if to_email['member_id']:
            to_member = aegis.model.Member.get_auth(to_email['member_id'], dbconn=dbconn)
            if to_member:
                email_data['to_email'] = to_email['email']
                if to_member.get('google_user', {}).get('name'):
                    email_data['to_name'] = to_member['google_user']['name']
                else:
                    email_data['to_name'] = ''
        kwargs = {}
        kwargs['domain'] = email_data['domain'] = dbconn.domain
        # Call an email template function if it has already been registered.
        if kwargs['domain'] in cls.template_registry and cls.template_registry[kwargs['domain']]:
            domain_templates = cls.template_registry[kwargs['domain']]
            fn = domain_templates[email_type['email_type_name']]
            email_data = fn(email_data)
        kwargs['reply_to'] = email_data['from_addr']
        email_data['nl2br'] = aegis.stdlib.nl2br
        email_data['format_integer'] = aegis.stdlib.format_integer
        # It's a mouthful to convert this to Pacific time
        # XXX TODO this should be on UTC, maybe without timezone information
        email_data['send_dttm_str'] = email_tracking['send_dttm'].astimezone(pytz.timezone('US/Pacific')).strftime('%b %d, %Y, %-H:%-M %p')
        email_data['options'] = options
        email_data['current_year'] = datetime.date.today().year
        email_data['email_tracking_id'] = email_tracking_id
        email_data['email_uuid'] = email_tracking['email_uuid']
        email_template = 'email/%s' % email_type['template_name']
        #kwargs['debug'] = True
        email_msg = cls.render_email(None, email_data['from_addr'], email_data['to_addr'], email_data['subject'], email_template, email_data, **kwargs)
        if email_msg:
            #logging.warning(email_msg)
            sent = cls.sendmail(from_email['email'], to_email['email'], email_msg)
            # Record email_tracking as sent
            if sent:
                email_tracking.mark_sent(dbconn=dbconn)
            return True
        return False


    @classmethod
    def sendmail(cls, from_addr, to_addrs, msg):
        if not options.smtp_host or not options.smtp_port:
            return logging.warning('No SMTP host or port supplied, not sending email')
        try:
            if options.smtp_user:
                conn = smtplib.SMTP_SSL(options.smtp_host, options.smtp_port)
                conn.login(options.smtp_user, options.smtp_pass)
            else:
                conn = smtplib.SMTP(options.smtp_host, options.smtp_port)
            conn.sendmail(from_addr, to_addrs, msg)
            logging.warning('Sent email to: %s' % to_addrs)
            return conn.quit()
        except Exception as ex:
            logging.exception(ex)



#-*- coding: utf-8 -*-
#
# Core common Data Model that will applies to a lot of applications


# Python Imports
import datetime
import hashlib
import json
import logging
import random
import uuid

# Project Imports
import aegis.stdlib
import aegis.database
import aegis.config

db = aegis.database.db


class SqlDiff(aegis.database.Row):
    table_name = 'sql_diff'
    id_column = 'sql_diff_id'

    @staticmethod
    def create_table():
        sql_diff_table = """
            CREATE TABLE IF NOT EXISTS
            sql_diff (
              sql_diff_id SERIAL NOT NULL,
              sql_diff_name VARCHAR(80) NOT NULL,
              create_dttm TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
              applied_dttm TIMESTAMP DEFAULT NULL,
              PRIMARY KEY (sql_diff_name)
            )"""
        return db().execute(sql_diff_table)

    @staticmethod
    def insert(sql_diff_name):
        sql = 'INSERT INTO sql_diff (sql_diff_name) VALUES (%s)'
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING sql_diff_id'
        return db().execute(sql, sql_diff_name)

    @classmethod
    def scan(cls):
        sql = 'SELECT * FROM sql_diff'
        return db().query(sql, cls=cls)

    @staticmethod
    def mark_applied(sql_diff_name):
        sql = 'UPDATE sql_diff SET applied_dttm=NOW() WHERE sql_diff_name=%s'
        return db().execute(sql, sql_diff_name)

    @classmethod
    def scan_unapplied(cls):
        sql = """SELECT * FROM sql_diff WHERE applied_dttm IS NULL ORDER BY SUBSTRING(sql_diff_name from 5 for 3) ASC"""
        return db().query(sql, cls=cls)


class UserAgent(aegis.database.Row):
    table_name = 'user_agent'
    id_column = 'user_agent_id'

    @staticmethod
    def insert(user_agent_tx):
        user_agent_md5 = hashlib.md5(str(user_agent_tx).encode('utf-8')).hexdigest()
        sql = 'INSERT INTO user_agent (user_agent_tx, user_agent_md5) VALUES (%s, %s)'
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING user_agent_id'
        return db().execute(sql, user_agent_tx, user_agent_md5)

    @classmethod
    def get_agent(cls, user_agent_tx):
        user_agent_md5 = hashlib.md5(str(user_agent_tx).encode('utf-8')).hexdigest()
        sql = 'SELECT * FROM user_agent WHERE user_agent_md5=%s'
        return db().get(sql, user_agent_md5, cls=cls)

    @classmethod
    def set_user_agent(cls, user_agent_tx):
        user_agent_row = cls.get_agent(user_agent_tx)
        if not user_agent_row:
            try:
                user_agent_id = cls.insert(user_agent_tx)
                user_agent_row = cls.get_id(user_agent_id)
            except (aegis.database.PgsqlIntegrityError, aegis.database.MysqlIntegrityError) as ex:
                user_agent_row = cls.get_agent(user_agent_tx)
                if not user_agent_row:
                    logging.exception(ex)
        return user_agent_row

    @staticmethod
    def set_robot_user_id(user_agent_id, robot_user_id):
        sql = 'UPDATE user_agent SET robot_user_id=%s WHERE user_agent_id=%s'
        return db().execute(sql, robot_user_id, user_agent_id)

    @staticmethod
    def set_robot_ind(user_agent_id, robot_ind):
        user_agent_id = int(user_agent_id)
        robot_ind = bool(robot_ind)
        if robot_ind not in [True, False]:
            return False
        sql = 'UPDATE user_agent SET robot_ind=%s WHERE user_agent_id=%s'
        return db().execute(sql, robot_ind, user_agent_id)

    def set_ua_json(self, ua_json):
        if self['user_agent_json'] or not ua_json:
            return
        sql = "UPDATE user_agent SET user_agent_json=%s WHERE user_agent_id=%s"
        return db().execute(sql, ua_json, self['user_agent_id'])


class User(aegis.database.Row):
    table_name = 'user_' if aegis.database.pgsql_available else 'user'
    table_names = {'pgsql': 'user_', 'mysql': 'user'}
    id_column = 'user_id'

    @classmethod
    def insert(cls, user_agent_id):
        sql = 'INSERT INTO '+cls.table_name+' (user_agent_id) VALUES (%s)'
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING user_id'
        return db().execute(sql, user_agent_id)

    def set_member_id(self, member_id):
        sql = 'UPDATE '+self.table_name+' SET member_id=%s WHERE user_id=%s'
        return db().execute(sql, member_id, self['user_id'])

    def set_preferences(self, preferences):
        preferences = json.dumps(preferences, cls=aegis.stdlib.DateTimeEncoder)
        sql = "UPDATE "+self.table_name+" SET preferences=%s WHERE user_id=%s"
        return db().execute(sql, preferences, self['user_id'])


class Email(aegis.database.Row):
    table_name = 'email'
    id_column = 'email_id'

    @staticmethod
    def insert(email):
        sql = "INSERT INTO email (email) VALUES (%s)"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING email_id'
        return db().execute(sql, email)

    @classmethod
    def set_email(cls, email):
        email_row = cls.get_email(email)
        if not email_row:
            try:
                email_id = cls.insert(email)
                email_row = cls.get_id(email_id)
            except (aegis.database.PgsqlUniqueViolation, aegis.database.MysqlIntegrityError) as ex:
                logging.error("Duplicate Key Error   aegis.model.Email.set_email(%s). Very unlikely. Read from DB and carry one.", email)
                email_row = cls.get_email(email)
        return email_row

    @classmethod
    def get_email(cls, email):
        sql = "SELECT * FROM email WHERE email=%s"
        return db().get(sql, email, cls=cls)

    @classmethod
    def get_member_id(cls, member_id):
        sql = "SELECT * FROM email WHERE member_id=%s"
        return db().get(sql, member_id, cls=cls)

    def set_member_id(self, member_id):
        sql = "UPDATE email SET member_id=%s WHERE email_id=%s"
        return db().execute(sql, member_id, self['email_id'])

    def set_google_user_id(self, google_user_id):
        sql = "UPDATE email SET google_user_id=%s WHERE email_id=%s"
        return db().execute(sql, google_user_id, self['email_id'])

    def deidentify(self, email_addr):
        # overwrite email with email_addr, stop referencing google_user.google_user_id, and mark deleted
        sql = "UPDATE email SET email=%s, google_user_id=NULL, delete_dttm=NOW() WHERE email_id=%s"
        return db().execute(sql, email_addr, self['email_id'])

    def mark_verified(self):
        sql = "UPDATE email SET verify_dttm=NOW() WHERE email_id=%s AND verify_dttm IS NULL"
        return db().execute(sql, self['email_id'])


class Member(aegis.database.Row):
    table_name = 'member'
    id_column = 'member_id'

    @staticmethod
    def insert(email_id):
        sql = "INSERT INTO member (email_id) VALUES (%s)"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING member_id'
        return db().execute(sql, email_id)

    @classmethod
    def set_member(cls, email_id, marketing_id=None, **kwargs):
        member = cls.get_email_id(email_id)
        if not member:
            columns = {'email_id': email_id}
            columns.update(kwargs)
            if marketing_id:
                columns['marketing_id'] = marketing_id
            member_id = cls.insert_columns(**columns)
            member = cls.get_id(member_id)
        return member

    @classmethod
    def get_email_id(cls, email_id):
        sql = "SELECT * FROM member WHERE email_id = %s"
        return db().get(sql, email_id, cls=cls)

    def set_google_user_id(self, google_user_id):
        sql = "UPDATE member SET google_user_id=%s WHERE member_id=%s"
        return db().execute(sql, google_user_id, self['member_id'])

    @classmethod
    def get_auth(cls, member_id, dbconn=None):
        if not dbconn:
            dbconn = db()
        if not member_id:
            return None
        member = cls.get_id(member_id, dbconn=dbconn)
        if not member:
            return None
        if member.get('email_id'):
            email = Email.get_id(member['email_id'], dbconn=dbconn)
            if email:
                member['email'] = email
        if member.get('google_user_id'):
            member['google_user'] = GoogleUser.get_id(member['google_user_id'], dbconn=dbconn)
            member['picture_url'] = member['google_user']['picture_url']
        return member

    def deidentify(self):
        # stop referencing google_user.google_user_id, and mark deleted
        sql = "UPDATE member SET google_user_id=NULL, delete_dttm=NOW() WHERE member_id=%s"
        return db().execute(sql, self['member_id'])

    def mark_verified(self):
        sql = "UPDATE member SET verify_dttm=NOW() WHERE member_id=%s AND verify_dttm IS NULL"
        return db().execute(sql, self['member_id'])


class MemberAuth(aegis.database.Row):
    table_name = 'member_auth'
    id_column = 'member_auth_id'
    data_columns = ('member_auth_type_id', 'magic_token', 'user_id', 'email_id', 'member_id', 'ip_address', 'expire_dttm', 'register_flag', 'login_flag')

    @classmethod
    def insert(cls, **columns):
        columns['magic_token'] = aegis.stdlib.magic_token()
        # XXX TODO Do input validation here
        return cls.insert_columns(**columns)

    @classmethod
    def get_auth(cls, member_id, member_auth_id, magic_token):
        sql = """
        SELECT *
          FROM member
          JOIN member_auth USING (member_id)
         WHERE member_id=%s
           AND member_auth_id=%s
           AND magic_token=%s
           AND member.delete_dttm IS NULL
           AND member_auth.expire_dttm > NOW()
           AND member_auth.delete_dttm IS NULL"""
        return db().get(sql, member_id, member_auth_id, magic_token, cls=cls)

    @classmethod
    def scan_member(cls, member_id):
        sql = """
        SELECT *
          FROM member
          JOIN member_auth USING (member_id)
         WHERE member_id=%s
           AND member.delete_dttm IS NULL
           AND member_auth.expire_dttm > NOW()
           AND member_auth.delete_dttm IS NULL"""
        return db().query(sql, member_id, cls=cls)

    def refresh(self, auth_duration_sec):
        auth_duration_sec = aegis.stdlib.validate_int(auth_duration_sec)
        sql = "UPDATE member_auth SET expire_dttm=NOW() + INTERVAL '%s SECOND' WHERE member_auth_id=%s"
        return db().execute(sql, auth_duration_sec, self['member_auth_id'])

    def revoke(self):
        sql = "UPDATE member_auth SET delete_dttm=NOW() WHERE member_auth_id=%s"
        return db().execute(sql, self['member_auth_id'])


class MemberAuthType(aegis.database.Row):
    table_name = 'member_auth_type'
    id_column = 'member_auth_type_id'

    @classmethod
    def get_name(cls, member_auth_type_name):
        sql = "SELECT * FROM member_auth_type WHERE member_auth_type_name=%s"
        return db().get(sql, member_auth_type_name, cls=cls)


class GoogleUser(aegis.database.Row):
    table_name = 'google_user'
    id_column = 'google_user_id'

    @staticmethod
    def insert(google_id, email_id, member_id, name, picture_url, email_verified):
        sql = "INSERT INTO google_user (google_id, email_id, member_id, name, picture_url, email_verified) VALUES (%s, %s, %s, %s, %s, %s) RETURNING google_user_id"
        return db().execute(sql, google_id, email_id, member_id, name, picture_url, email_verified)

    @classmethod
    def set_google_user(cls, google_id, email_id, member_id, name, picture_url, email_verified):
        google_user = cls.get_google_id(google_id)
        if not google_user:
            google_user_id = cls.insert(google_id, email_id, member_id, name, picture_url, email_verified)
            google_user = cls.get_id(google_user_id)
        return google_user

    @classmethod
    def get_google_id(cls, google_id):
        sql = "SELECT * FROM google_user WHERE google_id=%s"
        return db().get(sql, google_id, cls=cls)

    @classmethod
    def get_member_id(cls, member_id):
        sql = "SELECT * FROM google_user WHERE member_id=%s"
        return db().get(sql, member_id, cls=cls)

    def hard_delete(self):
        sql = "DELETE FROM google_user WHERE google_user_id=%s"
        return db().execute(sql, self['google_user_id'])

    def set_scopes(self, scopes):
        sql = "UPDATE google_user SET scopes=%s WHERE google_user_id=%s"
        return db().execute(sql, scopes, self['google_user_id'])


class GoogleAccess(aegis.database.Row):
    table_name = 'google_access'
    id_column = 'google_access_id'
    primary_key = ('google_access_id')

    @classmethod
    def get_google_user_id(cls, google_user_id):
        sql = "SELECT * FROM google_access WHERE google_user_id=%s ORDER BY expire_dttm DESC LIMIT 1"
        return db().get(sql, google_user_id, cls=cls)

    @classmethod
    def purge_expired(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "DELETE FROM google_access WHERE expire_dttm < NOW()"
        return dbconn.execute(sql)

    def hard_delete(self):
        sql = "DELETE FROM google_access WHERE google_access_id=%s"
        return db().execute(sql, self['google_access_id'])

    @classmethod
    def hard_delete_google_user(cls, google_user_id):
        sql = "DELETE FROM google_access WHERE google_user_id=%s"
        return db().execute(sql, google_user_id)


class GooglePicture(aegis.database.Row):
    table_name = 'google_picture'
    id_column = 'google_picture_id'
    primary_key = ('google_picture_id')

    @classmethod
    def get_google_user_id(cls, google_user_id, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM google_picture WHERE google_user_id=%s"
        return dbconn.get(sql, google_user_id, cls=cls)

    def hard_delete(self):
        sql = "DELETE FROM google_picture WHERE google_picture_id=%s"
        return db().execute(sql, self['google_picture_id'])

    @classmethod
    def hard_delete_google_user(cls, google_user_id):
        sql = "DELETE FROM google_picture WHERE google_user_id=%s"
        return db().execute(sql, google_user_id)


class EmailType(aegis.database.Row):
    table_name = 'email_type'
    id_column = 'email_type_id'

    @classmethod
    def scan(cls):
        sql = "SELECT * FROM email_type"
        return db().query(sql, cls=cls)

    @classmethod
    def get_name(cls, email_type_name):
        sql = "SELECT * FROM email_type WHERE email_type_name=%s"
        return db().get(sql, email_type_name, cls=cls)


class EmailTracking(aegis.database.Row):
    table_name = 'email_tracking'
    id_column = 'email_tracking_id'

    @staticmethod
    def insert(email_type_id, from_email_id, to_email_id, email_data):
        sql = "INSERT INTO email_tracking (email_type_id, from_email_id, to_email_id, email_uuid, email_data, send_dttm) VALUES (%s, %s, %s, %s, %s, NOW())"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING email_tracking_id'
        return db().execute(sql, email_type_id, from_email_id, to_email_id, uuid.uuid4().hex, email_data)

    @classmethod
    def get_params(cls, email_tracking_id, email_uuid, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM email_tracking WHERE email_tracking_id=%s AND email_uuid=%s"
        return dbconn.get(sql, email_tracking_id, email_uuid, cls=cls)

    @classmethod
    def scan_mailer(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM email_tracking WHERE send_dttm < NOW() AND sent_dttm IS NULL AND delete_dttm IS NULL AND claimed_dttm IS NULL"
        return dbconn.query(sql, cls=cls)

    def claim(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'UPDATE email_tracking SET claimed_dttm=NOW() WHERE email_tracking_id=%s AND claimed_dttm IS NULL'
        return dbconn.execute(sql, self['email_tracking_id'])

    @staticmethod
    def clear_claims(minutes=1, dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "UPDATE email_tracking SET claimed_dttm=NULL WHERE claimed_dttm < NOW() - INTERVAL '%s MINUTE' AND sent_dttm IS NULL" % int(minutes)
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "UPDATE email_tracking SET claimed_dttm=NULL WHERE claimed_dttm < NOW() - INTERVAL %s MINUTE AND sent_dttm IS NULL" % int(minutes)
        return dbconn.execute_rowcount(sql)

    def mark_sent(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE email_tracking SET sent_dttm=NOW() WHERE email_tracking_id=%s AND sent_dttm IS NULL"
        return dbconn.execute(sql, self['email_tracking_id'])

    def mark_delivered(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE email_tracking SET deliver_dttm=NOW() WHERE email_tracking_id=%s AND deliver_dttm IS NULL"
        return dbconn.execute(sql, self['email_tracking_id'])

    def mark_opened(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE email_tracking SET open_dttm=NOW() WHERE email_tracking_id=%s AND open_dttm IS NULL"
        return dbconn.execute(sql, self['email_tracking_id'])

    def mark_clicked(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE email_tracking SET click_dttm=NOW() WHERE email_tracking_id=%s AND click_dttm IS NULL"
        return dbconn.execute(sql, self['email_tracking_id'])

    def mark_deleted(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE email_tracking SET delete_dttm=NOW() WHERE email_tracking_id=%s AND delete_dttm IS NULL"
        return dbconn.execute(sql, self['email_tracking_id'])


class EmailLink(aegis.database.Row):
    table_name = 'email_link'
    id_column = 'email_link_id'

    @staticmethod
    def insert(email_id):
        sql = "INSERT INTO email_link (email_id, magic_token) VALUES (%s, %s)"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING email_link_id'
        return db().execute(sql, email_id, aegis.stdlib.magic_token())

    def mark_accessed(self):
        sql = "UPDATE email_link SET access_dttm=NOW(), delete_dttm=NOW() WHERE email_link_id=%s AND access_dttm IS NULL"
        return db().execute(sql, self['email_link_id'])

    def set_email_tracking_id(self, email_tracking_id):
        sql = "UPDATE email_link SET email_tracking_id=%s WHERE email_link_id=%s AND email_tracking_id IS NULL"
        return db().execute(sql, email_tracking_id, self['email_link_id'])

    @classmethod
    def scan_latest(cls, email_id, limit=6):
        sql = "SELECT * FROM email_tracking JOIN email_link USING (email_tracking_id) WHERE email_type_id=2 AND to_email_id=%s ORDER BY email_tracking_id DESC LIMIT %s"
        return db().query(sql, email_id, limit, cls=cls)

    @classmethod
    def get_token(cls, magic_token):
        sql = "SELECT * FROM email_link WHERE magic_token=%s"
        return db().get(sql, magic_token, cls=cls)

    @classmethod
    def get_id_token(cls, email_link_id, magic_token):
        sql = "SELECT * FROM email_link WHERE email_link_id=%s AND magic_token=%s"
        return db().get(sql, email_link_id, magic_token, cls=cls)


class Pageview(aegis.database.Row):
    table_name = 'pageview'
    id_column = 'pageview_id'

    @classmethod
    def insert(cls, user_id, member_id, url_path, url_query, url_tx, request_name):
        sql = "INSERT INTO pageview (user_id, member_id, url_path, url_query, url_tx, request_name) VALUES (%s, %s, %s, %s, %s, %s)"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING pageview_id'
        return db().execute(sql, user_id, member_id, url_path, url_query, url_tx, request_name)


class HydraType(aegis.database.Row):
    table_name = 'hydra_type'
    id_column = 'hydra_type_id'
    data_columns = ('hydra_type_name', 'hydra_type_desc', 'priority_ndx', 'next_run_sql', 'claimed_dttm', 'run_host', 'run_env')

    @classmethod
    def get_name(cls, hydra_type_name, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM hydra_type WHERE hydra_type_name=%s"
        return dbconn.get(sql, hydra_type_name, cls=cls)

    @classmethod
    def scan(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM hydra_type ORDER BY next_run_dttm ASC, status ASC, priority_ndx ASC"
        return dbconn.query(sql, cls=cls)

    def run_now(self):
        sql = "UPDATE hydra_type SET next_run_dttm=NOW(), status='live', claimed_dttm=NULL WHERE hydra_type_id=%s"
        return db().execute(sql, self['hydra_type_id'])

    def set_status(self, status, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE hydra_type SET status=%s WHERE hydra_type_id=%s"
        return dbconn.execute(sql, status, self['hydra_type_id'])

    @classmethod
    def get_runnable(cls, hydra_type_id, env, dbconn=None):
        sql = """SELECT hydra_type_id, hydra_type_name, next_run_sql
                   FROM hydra_type
                  WHERE next_run_dttm <= NOW()
                    AND status='live'
                    AND hydra_type_id=%s
                    AND run_env=%s
                    AND claimed_dttm IS NULL
                    AND next_run_sql IS NOT NULL"""
        if not dbconn:
            dbconn = db()
        return dbconn.get(sql, hydra_type_id, env, cls=cls)

    def schedule_next(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = """
            UPDATE hydra_type
               SET run_cnt=run_cnt+1,
                   last_run_dttm=next_run_dttm,
                   claimed_dttm=NULL,
                   next_run_dttm="""+self['next_run_sql']+"""
             WHERE hydra_type_id=%s
               AND status = 'live'"""
        return dbconn.execute(sql, self['hydra_type_id'])

    @staticmethod
    def clear_running(dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "UPDATE hydra_type SET status='live' WHERE status='running' and next_run_dttm < NOW() - INTERVAL '45 MINUTE'"
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "UPDATE hydra_type SET status='live' WHERE status='running' and next_run_dttm < NOW() - INTERVAL 45 MINUTE"
        return dbconn.execute(sql)

    def claim(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'UPDATE hydra_type SET claimed_dttm=NOW() WHERE hydra_type_id=%s AND claimed_dttm IS NULL'
        return dbconn.execute(sql, self['hydra_type_id'])

    def unclaim(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'UPDATE hydra_type SET claimed_dttm=NULL WHERE hydra_type_id=%s AND claimed_dttm IS NOT NULL'
        return dbconn.execute(sql, self['hydra_type_id'])

    @classmethod
    def past_items(cls, minutes=5, dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "SELECT * FROM hydra_type WHERE claimed_dttm < NOW() - INTERVAL '%s MINUTE' AND claimed_dttm > next_run_dttm" % int(minutes)
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "SELECT * FROM hydra_type WHERE claimed_dttm < NOW() - INTERVAL %s MINUTE AND claimed_dttm > next_run_dttm" % int(minutes)
        return dbconn.query(sql, cls=cls)

    @staticmethod
    def clear_claims(minutes=5, dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "UPDATE hydra_type SET claimed_dttm=NULL WHERE claimed_dttm < NOW() - INTERVAL '%s MINUTE' AND status='live'" % int(minutes)
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "UPDATE hydra_type SET claimed_dttm=NULL WHERE claimed_dttm < NOW() - INTERVAL %s MINUTE AND status='live'" % int(minutes)
        return dbconn.execute(sql)


class HydraQueue(aegis.database.Row):
    table_name = 'hydra_queue'
    id_column = 'hydra_queue_id'
    data_columns = ('hydra_type_id', 'priority_ndx', 'work_host', 'work_env', 'work_data', 'work_dttm', 'start_dttm', 'claimed_dttm', 'finish_dttm', 'try_cnt', 'error_cnt')

    @classmethod
    def scan_work_priority(cls, limit=10, hostname=None, env=None, dbconn=None):
        sql = """
        SELECT hydra_queue.*,
               hydra_type.hydra_type_name,
               hydra_type.hydra_type_desc
          FROM hydra_queue
          JOIN hydra_type USING (hydra_type_id)
         WHERE hydra_queue.work_dttm <= NOW()
           AND (hydra_queue.work_host=%s OR hydra_queue.work_host IS NULL)
           AND (hydra_queue.work_env=%s OR hydra_queue.work_env IS NULL)
           AND hydra_queue.claimed_dttm IS NULL
           AND hydra_queue.finish_dttm IS NULL
           AND hydra_queue.delete_dttm IS NULL
           AND hydra_type.status <> 'paused'
      ORDER BY hydra_queue.priority_ndx ASC
         LIMIT %s"""
        if not dbconn:
            dbconn = db()
        return dbconn.query(sql, hostname, env, limit, cls=cls)

    @classmethod
    def scan_work(cls, limit=10, hostname=None, env=None, dbconn=None):
        sql = """
        SELECT hydra_queue.*,
               hydra_type.hydra_type_name,
               hydra_type.hydra_type_desc
          FROM hydra_queue
          JOIN hydra_type USING (hydra_type_id)
         WHERE hydra_queue.work_dttm <= NOW()
           AND (hydra_queue.work_host=%s OR hydra_queue.work_host IS NULL)
           AND (hydra_queue.work_env=%s OR hydra_queue.work_env IS NULL)
           AND hydra_queue.claimed_dttm IS NULL
           AND hydra_queue.finish_dttm IS NULL
           AND hydra_queue.delete_dttm IS NULL
           AND hydra_type.status <> 'paused'
      ORDER BY hydra_queue.work_dttm ASC
         LIMIT %s"""
        if not dbconn:
            dbconn = db()
        return dbconn.query(sql, hostname, env, limit, cls=cls)

    @classmethod
    def scan(cls, limit=100):
        # XXX TODO using work_host, work_env
        if type(db()) is aegis.database.PostgresConnection:
            sql = "SELECT hydra_queue.*, hydra_type.hydra_type_name, hydra_type.next_run_sql, hydra_type.status FROM hydra_queue JOIN hydra_type USING (hydra_type_id) WHERE finish_dttm IS NULL AND hydra_queue.delete_dttm IS NULL ORDER BY claimed_dttm DESC NULLS LAST, hydra_queue.work_dttm ASC LIMIT %s"
        elif type(db()) is aegis.database.MysqlConnection:
            sql = "SELECT hydra_queue.*, hydra_type.hydra_type_name, hydra_type.next_run_sql, hydra_type.status FROM hydra_queue JOIN hydra_type USING (hydra_type_id) WHERE finish_dttm IS NULL AND hydra_queue.delete_dttm IS NULL ORDER BY claimed_dttm DESC, hydra_queue.work_dttm ASC LIMIT %s"
        return db().query(sql, limit, cls=cls)

    @classmethod
    def scan_work_type(cls, hydra_type_id):
        # XXX TODO using work_host, work_env
        sql = "SELECT * FROM hydra_queue WHERE work_dttm <= NOW() AND claimed_dttm IS NULL AND finish_dttm IS NULL AND delete_dttm IS NULL AND hydra_type_id=%s"
        return db().query(sql, hydra_type_id, cls=cls)

    @classmethod
    def scan_existing(cls, hydra_type_id, data):
        # XXX TODO using work_host, work_env
        sql = "SELECT * FROM hydra_queue WHERE finish_dttm IS NULL AND hydra_type_id=%s AND work_data=%s"
        return db().query(sql, hydra_type_id, data, cls=cls)

    @classmethod
    def count_live(cls):
        sql = "SELECT COUNT(*) AS queue_cnt FROM hydra_queue WHERE finish_dttm IS NULL AND delete_dttm IS NULL"
        return db().get(sql, cls=cls)

    @classmethod
    def scan_type(cls, hydra_type_id):
        sql = "SELECT * FROM hydra_queue WHERE finish_dttm IS NULL AND delete_dttm IS NULL AND hydra_type_id=%s"
        return db().query(sql, hydra_type_id, cls=cls)

    def claim(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'UPDATE hydra_queue SET claimed_dttm=NOW() WHERE hydra_queue_id=%s AND claimed_dttm IS NULL'
        return dbconn.execute(sql, self['hydra_queue_id'])

    def unclaim(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'UPDATE hydra_queue SET claimed_dttm=NULL WHERE hydra_queue_id=%s AND claimed_dttm IS NOT NULL'
        return dbconn.execute(sql, self['hydra_queue_id'])

    def incr_try_cnt(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        self['try_cnt'] += 1
        sql = 'UPDATE hydra_queue SET try_cnt=try_cnt+1 WHERE hydra_queue_id=%s'
        return dbconn.execute(sql, self['hydra_queue_id'])

    def incr_error_cnt(self, minutes=1, dbconn=None):
        if not dbconn:
            dbconn = db()
        self['error_cnt'] += 1
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "UPDATE hydra_queue SET error_cnt=error_cnt+1, start_dttm=NULL, work_dttm=NOW() + INTERVAL '%s MINUTE' WHERE hydra_queue_id=%s"
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "UPDATE hydra_queue SET error_cnt=error_cnt+1, start_dttm=NULL, work_dttm=NOW() + INTERVAL %s MINUTE WHERE hydra_queue_id=%s"
        return dbconn.execute(sql, minutes, self['hydra_queue_id'])

    def start(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        hydra_type = HydraType.get_id(self['hydra_type_id'], dbconn=dbconn)
        if hydra_type['next_run_sql']:
            hydra_type.set_status('running', dbconn=dbconn)
        sql = 'UPDATE hydra_queue SET start_dttm=NOW() WHERE hydra_queue_id=%s'
        return dbconn.execute(sql, self['hydra_queue_id'])

    def complete(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        hydra_type = HydraType.get_id(self['hydra_type_id'], dbconn=dbconn)
        if hydra_type['next_run_sql'] and hydra_type['status'] == 'running':
            hydra_type.set_status('live', dbconn=dbconn)
        sql = 'UPDATE hydra_queue SET finish_dttm=NOW() WHERE hydra_queue_id=%s'
        return dbconn.execute(sql, self['hydra_queue_id'])

    def finish(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'UPDATE hydra_queue SET finish_dttm=NOW() WHERE hydra_queue_id=%s'
        return dbconn.execute(sql, self['hydra_queue_id'])

    @staticmethod
    def purge_completed(dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "DELETE FROM hydra_queue WHERE finish_dttm IS NOT NULL AND finish_dttm < NOW()"
        return dbconn.execute(sql)

    @staticmethod
    def clear_claims(minutes=15, dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "UPDATE hydra_queue SET claimed_dttm=NULL, start_dttm=NULL WHERE claimed_dttm < NOW() - INTERVAL '%s MINUTE' AND finish_dttm IS NULL" % int(minutes)
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "UPDATE hydra_queue SET claimed_dttm=NULL, start_dttm=NULL WHERE claimed_dttm < NOW() - INTERVAL %s MINUTE AND finish_dttm IS NULL" % int(minutes)
        return dbconn.execute(sql)

    @classmethod
    def past_items(cls, minutes=15, dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "SELECT * FROM hydra_queue WHERE work_dttm < NOW() - INTERVAL '%s MINUTE' ORDER BY work_dttm ASC LIMIT 50" % int(minutes)
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "SELECT * FROM hydra_queue WHERE work_dttm < NOW() - INTERVAL %s MINUTE ORDER BY work_dttm ASC LIMIT 50" % int(minutes)
        return dbconn.query(sql, cls=cls)

    def run_now(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE hydra_queue SET work_dttm=NOW(), claimed_dttm=NULL WHERE hydra_queue_id=%s"
        return dbconn.execute(sql, self['hydra_queue_id'])

    def singleton(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM hydra_queue WHERE hydra_type_id=%s AND hydra_queue_id <> %s AND claimed_dttm IS NOT NULL AND finish_dttm IS NULL AND delete_dttm IS NULL"
        return dbconn.query(sql, self['hydra_type_id'], self['hydra_queue_id'])


class ReportType(aegis.database.Row):
    table_name = 'report_type'
    id_column = 'report_type_id'

    @classmethod
    def insert(cls, report_type_name, report_sql):
        sql = "INSERT INTO report_type (report_type_name, report_sql) VALUES (%s, %s)"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING report_type_id'
        return db().execute(sql, report_type_name, report_sql)

    @classmethod
    def scan(cls):
        sql = "SELECT * FROM report_type"
        return db().query(sql, cls=cls)

    @staticmethod
    def set_name(report_type_id, report_type_name):
        sql = 'UPDATE report_type SET report_type_name=%s WHERE report_type_id=%s'
        return db().execute(sql, report_type_name, report_type_id)

    @staticmethod
    def set_sql(report_type_id, report_type_sql):
        sql = 'UPDATE report_type SET report_sql=%s WHERE report_type_id=%s'
        return db().execute(sql, report_type_sql, report_type_id)


class Build(aegis.database.Row):
    table_name = 'build'
    id_column = 'build_id'

    @classmethod
    def scan(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM build ORDER BY build_id DESC"
        return dbconn.query(sql, cls=cls)

    def set_output(self, build_step, output_tx, exit_status=None, dbconn=None):
        if not dbconn:
            dbconn = db()
        # We have to be super explicit because MySQL CONCAT returns NULL if any param is NULL, even though Postgres ignores NULL and concatenates the other arguments
        if build_step == 'build':
            if self['build_output_tx'] is not None:
                sql = "UPDATE build SET build_output_tx=CONCAT(build_output_tx, %s), build_exit_status=%s WHERE build_id=%s AND build_output_tx IS NOT NULL"
            else:
                sql = "UPDATE build SET build_output_tx=%s, build_exit_status=%s WHERE build_id=%s AND build_output_tx IS NULL"
        elif build_step == 'deploy':
            if self['deploy_output_tx'] is not None:
                sql = "UPDATE build SET deploy_output_tx=CONCAT(deploy_output_tx, %s), deploy_exit_status=%s WHERE build_id=%s AND deploy_output_tx IS NOT NULL"
            else:
                sql = "UPDATE build SET deploy_output_tx=%s, deploy_exit_status=%s WHERE build_id=%s AND deploy_output_tx IS NULL"
        elif build_step == 'revert':
            if self['revert_output_tx'] is not None:
                sql = "UPDATE build SET revert_output_tx=CONCAT(revert_output_tx, %s), revert_exit_status=%s WHERE build_id=%s AND revert_output_tx IS NOT NULL"
            else:
                sql = "UPDATE build SET revert_output_tx=%s, revert_exit_status=%s WHERE build_id=%s AND revert_output_tx IS NULL"
        else:
            aegis.stdlib.logw(build_step, "BUILD_STEP NOT IN LIST")
            return
        return dbconn.execute(sql, output_tx, exit_status, self['build_id'])

    def set_version(self, version, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET version=%s WHERE build_id=%s AND version IS NULL"
        return dbconn.execute(sql, version, self['build_id'])

    def set_previous_version(self, previous_version, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET previous_version=%s WHERE build_id=%s AND previous_version IS NULL"
        return dbconn.execute(sql, previous_version, self['build_id'])

    def set_revision(self, revision, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET revision=%s WHERE build_id=%s AND revision IS NULL OR revision = 'HEAD'"
        return dbconn.execute(sql, revision, self['build_id'])

    def set_build_size(self, build_size, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET build_size=%s WHERE build_id=%s AND build_size IS NULL"
        return dbconn.execute(sql, build_size, self['build_id'])

    def set_message(self, message, build_step, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET " + build_step + "_message=%s WHERE build_id=%s AND " + build_step + "_message IS NULL"
        return dbconn.execute(sql, message, self['build_id'])

    def set_deployed(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET deploy_dttm=NOW() WHERE build_id=%s AND deploy_dttm IS NULL"
        return dbconn.execute(sql, self['build_id'])

    def set_reverted(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET revert_dttm=NOW() WHERE build_id=%s AND revert_dttm IS NULL"
        return dbconn.execute(sql, self['build_id'])

    def set_soft_deleted(self, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET delete_dttm=NOW() WHERE build_id=%s AND delete_dttm IS NULL"
        return dbconn.execute(sql, self['build_id'])

    def set_build_exec_sec(self, build_exec_sec, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "UPDATE build SET build_exec_sec=%s WHERE build_id=%s AND build_exec_sec IS NULL"
        return dbconn.execute(sql, build_exec_sec, self['build_id'])

    @classmethod
    def get_live_build(cls, env, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM build WHERE deploy_dttm IS NOT NULL AND deploy_exit_status=0 AND revert_dttm IS NULL AND env=%s AND build_target <> 'admin' ORDER BY deploy_dttm DESC LIMIT 1"
        return dbconn.get(sql, env, cls=cls)

    @classmethod
    def get_version(cls, version, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM build WHERE version=%s"
        return dbconn.get(sql, version, cls=cls)

    @classmethod
    def scan_dead_builds(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        if type(dbconn) is aegis.database.PostgresConnection:
            sql = "SELECT * FROM build WHERE delete_dttm IS NOT NULL OR (deploy_dttm IS NULL AND create_dttm < NOW() - INTERVAL '1 WEEK')"
        elif type(dbconn) is aegis.database.MysqlConnection:
            sql = "SELECT * FROM build WHERE delete_dttm IS NOT NULL OR (deploy_dttm IS NULL AND create_dttm < NOW() - INTERVAL 1 WEEK)"
        return dbconn.query(sql, cls=cls)

    @classmethod
    def scan_stale_builds(cls, env, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT * FROM build WHERE env=%s AND deploy_dttm IS NOT NULL AND delete_dttm IS NULL ORDER BY deploy_dttm DESC"
        return dbconn.query(sql, env, cls=cls)

    @classmethod
    def deployed_envs(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = "SELECT DISTINCT env FROM build WHERE delete_dttm IS NULL"
        return dbconn.query(sql, cls=cls)


class Cache(aegis.database.Row):
    table_name = 'cache'
    id_column = 'cache_id'

    # External Interface for simplified usage
    @classmethod
    def get_cache(cls, cache_key):
        cls.purge_expired()
        cache_obj = cls.get_key(cache_key)
        if cache_obj:
            return json.loads(cache_obj['cache_json'])

    @classmethod
    def set_cache(cls, cache_key, cache_obj, duration_s):
        cache_json = json.dumps(cache_obj, cls=aegis.stdlib.DateTimeEncoder)
        cache_expiry = datetime.datetime.utcnow() + datetime.timedelta(seconds=duration_s) + datetime.timedelta(seconds=random.randint(0, duration_s))
        cache_id = cls.set_key(cache_key, cache_json, cache_expiry)
        return cls.get_cache(cache_key)

    @staticmethod
    def purge_all():
        sql = "DELETE FROM cache"
        return db().execute(sql)

    # Internal Interface
    @staticmethod
    def insert(cache_key, cache_json, cache_expiry):
        sql = "INSERT INTO cache (cache_key, cache_json, cache_expiry) VALUES (%s, %s, %s)"
        if type(db()) is aegis.database.PostgresConnection:
            sql += ' RETURNING cache_id'
        try:
            return db().execute(sql, cache_key, cache_json, cache_expiry)
        except aegis.database.PgsqlUniqueViolation as ex:
            logging.warning("Ignoring duplicate key in cache")
        except aegis.database.MysqlIntegrityError as ex:
            logging.warning("Ignoring duplicate key in cache")

    @classmethod
    def get_key(cls, cache_key):
        sql = "SELECT * FROM cache WHERE cache_key=%s"
        return db().get(sql, cache_key, cls=cls)

    @staticmethod
    def update_key(cache_key, cache_json, cache_expiry):
        sql = "UPDATE cache SET cache_json=%s, cache_expiry=%s WHERE cache_key=%s"
        return db().execute(sql, cache_json, cache_expiry, cache_key)

    @classmethod
    def set_key(cls, cache_key, cache_json, cache_expiry):
        cache_obj = cls.get_key(cache_key)
        if cache_obj:
            cls.update_key(cache_key, cache_json, cache_expiry)
            cache_obj = cls.get_key(cache_key)
        else:
            cls.insert(cache_key, cache_json, cache_expiry)
            cache_obj = cls.get_key(cache_key)
        return cache_obj

    @staticmethod
    def del_key(cache_key):
        sql = "DELETE FROM cache WHERE cache_key=%s"
        return db().execute(sql, cache_key)

    @staticmethod
    def purge_expired():
        sql = "DELETE FROM cache WHERE cache_expiry < NOW()"
        return db().execute(sql)


class AuditSession(aegis.database.Row):
    table_name = 'audit_session'
    id_column = 'audit_session_id'


class AuditRequest(aegis.database.Row):
    table_name = 'audit_request'
    id_column = 'audit_request_id'

class AuditRequestData(aegis.database.Row):
    table_name = 'audit_request_data'
    id_column = 'audit_request_data_id'

    @classmethod
    def get_audit_request_id(cls, audit_request_id):
        sql = "SELECT * FROM audit_request_data WHERE audit_request_id=%s"
        return db().get(sql, audit_request_id, cls=cls)


class Monitor(aegis.database.Row):
    table_name = 'monitor'
    id_column = 'monitor_id'

    @classmethod
    def get_host_cmd(cls, host, cmd):
        sql = "SELECT * FROM monitor WHERE monitor_host=%s AND monitor_cmd=%s ORDER BY monitor_id DESC LIMIT 1"
        return db().get(sql, host, cmd, cls=cls)


class Marketing(aegis.database.Row):
    table_name = 'marketing'
    id_column = 'marketing_id'

    @classmethod
    def get_name(cls, name):
        sql = 'SELECT * FROM marketing WHERE marketing_name=%s'
        return db().get(sql, name, cls=cls)



#-*- coding: utf-8 -*-
#
# Fork of Tornado Database using Postgres and Mysql


# Python Imports
import logging
import os
import threading
import time

# Extern Imports
import tornado.options
from tornado.options import options

# Project Imports
import aegis.stdlib


# Import drivers as needed and set up error classes
pgsql_available = False
PgsqlIntegrityError = None
PgsqlOperationalError = None
PgsqlDatabaseError = None
PgsqlProgrammingError = None
PgsqlUniqueViolation = None
PgsqlAdminShutdown = None
try:
    import psycopg2
    pgsql_available = True
    # These are here for mapping errors from psycopg2 into application namespace
    PgsqlIntegrityError = psycopg2.IntegrityError
    PgsqlOperationalError = psycopg2.OperationalError
    PgsqlDatabaseError = psycopg2.Error
    PgsqlProgrammingError = psycopg2.ProgrammingError
    PgsqlUniqueViolation = psycopg2.errors.UniqueViolation
    PgsqlAdminShutdown = psycopg2.errors.AdminShutdown
except Exception as ex:
    #logging.error("Couldn't import psycopg2 - maybe that's ok for now - but shim the exception types.")
    #logging.exception(ex)
    class PgsqlIntegrityError(BaseException):
        pass
    class PgsqlOperationalError(BaseException):
        pass
    class PgsqlDatabaseError(BaseException):
        pass
    class PgsqlProgrammingError(BaseException):
        pass
    class PgsqlUniqueViolation(BaseException):
        pass
    class PgsqlAdminShutdown(BaseException):
        pass

mysql_available = False
MysqlIntegrityError = None
MysqlOperationalError = None
MysqlInterfaceError = None
MysqlDataError = None
try:
    import MySQLdb
    mysql_available = True
    # These are here for mapping errors from MySQLdb into application namespace
    MysqlIntegrityError = MySQLdb._exceptions.IntegrityError
    MysqlOperationalError = MySQLdb._exceptions.OperationalError
    MysqlInterfaceError = MySQLdb._exceptions.InterfaceError
    MysqlDataError = MySQLdb._exceptions.DataError
except Exception as ex:
    #logging.error("Couldn't import MySQLdb - maybe that's ok for now - but shim the exception types.")
    #logging.exception(ex)
    class MysqlIntegrityError(BaseException):
        pass
    class MysqlOperationalError(BaseException):
        pass
    class MysqlInterfaceError(BaseException):
        pass
    class MysqlDataErrorError(BaseException):
        pass


# Thread-safe persistent database connection
dbconns = threading.local()


def db(use_schema=None, autocommit=True, **kwargs):
    if not hasattr(dbconns, 'databases'):
        dbconns.databases = {}
    # psycopg and probably mysqldb are thread-safe but not multiprocess-safe. If it's a separate process, create a new connection.
    # You'll want to do a db().close() after the multiprocessing function returns since the SSL breaks down anyways
    pid = os.getpid()
    dbconns.databases.setdefault(pid, {})
    if pgsql_available:
        # Autocommit==False will be its own short-lived connection, not cached or pooled, so we don't end up with transactions open from other cursors.
        if not autocommit:
            return PostgresConnection.connect(autocommit=False)
        # Autocommit==True we can cache the database connection and let autocommit handle cursor-transaction-safety
        pg_database = kwargs.get('pg_database') or options.pg_database
        if pg_database not in dbconns.databases[pid]:
            dbconns.databases[pid][pg_database] = PostgresConnection.connect(**kwargs)
        if not use_schema:
            use_schema = pg_database
    if mysql_available:
        mysql_database = kwargs.get('mysql_database') or options.mysql_schema
        if mysql_database not in dbconns.databases[pid]:
            dbconns.databases[pid][mysql_database] = MysqlConnection.connect(**kwargs)
        if not use_schema:
            use_schema = mysql_database
    # Default situation - much better to be explicit which database we're connecting to!
    if not use_schema and len(dbconns.databases[pid]) == 1:
        use_schema = [dbconn for dbconn in dbconns.databases[pid].keys()][0]
    return dbconns.databases[pid][use_schema]


def dbnow(use_schema=None, dbconn=None):
    if not dbconn:
        dbconn = db(use_schema)
    return dbconn.get("SELECT NOW() AS now")

def sql_in_format(lst, cast):
    lst = [cast(lst_item) for lst_item in lst]
    format_strings = ','.join(['%s'] * len(lst))
    return lst, format_strings


class PostgresConnection(object):
    threads = {}

    def __init__(self, hostname, port, database, username=None, password=None, autocommit=True):
        """ Called by connect() class method using cls() notation """
        self.hostname = hostname
        self.port = port
        self.database = database
        args = "port={0} dbname={1}".format(self.port, self.database)
        if hostname is not None:
            args += " host={0}".format(hostname)
        if username is not None:
            args += " user={0}".format(username)
        if password is not None:
            args += " password={0}".format(password)
        self._db = None
        self._db_args = args
        self._autocommit = autocommit
        self._txn = None
        try:
            self._connect(autocommit=autocommit)
            if not autocommit:
                self._cursor()
        except Exception:
            logging.error("Cannot connect to PostgreSQL: %s", self.hostname, exc_info=True)

    @classmethod
    def connect(cls, autocommit=True, **kwargs):
        if 'pg_database' in kwargs:
            database = kwargs['pg_database']
            hostname = kwargs['pg_hostname']
            username = kwargs['pg_username']
            password = kwargs['pg_password']
            port = kwargs.get('pg_port', 5432)
        else:
            database = options.pg_database
            hostname = options.pg_hostname
            username = options.pg_username
            password = options.pg_password
            port = options.pg_port
        # force a new connection with a flag, or implicitly when using autocommit False
        if kwargs.get('force', False) or autocommit == False:
            return cls(hostname, port, database, username, password, autocommit=autocommit)
        # check existing connections
        ident = threading.current_thread().ident
        connections = cls.threads.setdefault(ident, {})
        if not database in connections:
            conn = cls(hostname, port, database, username, password)
            conn.database = database
            conn._connect(autocommit)
            cls.threads[ident][database] = conn
        return connections[database]

    def _connect(self, autocommit):
        self._db = psycopg2.connect(self._db_args)
        self._db.autocommit = autocommit

    def _cursor(self):
        # Connect if not connected.
        if self._db is None:
            self._connect(self._autocommit)
        # If auto-commit then return a new cursor immediately.
        if self._autocommit:
            return self._db.cursor()
        # The connection is single-use when doing a transaction via autocommit=False. That's the cursor to use until we commit the transaction.
        if not self._txn:
            self._txn = self._db.cursor()
        return self._txn

    def __del__(self):
        self.close()

    def close(self):
        if getattr(self, "_db", None) is not None:
            self._db.close()
            self._db = None

    def commit(self):
        if getattr(self, "_db", None) is not None:
            self._db.commit()
            self._txn = None

    def rollback(self):
        if getattr(self, "_db", None) is not None:
            self._db.rollback()
            self._txn = None

    def _execute(self, cursor, query, parameters, **kwargs):
        try:
            aegis.stdlib.incr_start(aegis.stdlib.get_timer(), 'database')
            cursor.execute(query, parameters)
            aegis.stdlib.incr_stop(aegis.stdlib.get_timer(), 'database')
            return
        except PgsqlUniqueViolation as ex:
            # UniqueViolation doesn't need to close connection, it needs to be handled in application
            raise
        except (psycopg2.Error, PgsqlAdminShutdown, PgsqlOperationalError) as ex:
            logging.error("General Error at PostgreSQL - close connection/rollback")
            logging.exception(ex)
            self.close()
            raise

    def iter(self, query, *parameters, **kwargs):
        """Returns an iterator for the given query and parameters."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters)
            column_names = [d[0] for d in cursor.description]
            if kwargs.get('cls'):
                for row in cursor:
                    yield kwargs['cls'](zip(column_names, row))
            else:
                for row in cursor:
                    yield Row(zip(column_names, row))
        finally:
            cursor.close()

    def query(self, query, *parameters, **kwargs):
        """ Returns a row list for the given query and parameters."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, **kwargs)
            column_names = [d[0] for d in cursor.description]
            cls = kwargs.get('cls')
            if cls:
                rows = [cls(list(zip(column_names, row))) for row in cursor]
            else:
                rows = [Row(zip(column_names, row)) for row in cursor]
            if kwargs.get('return_column_names'):
                return (rows, column_names)
            return rows
        finally:
            if not self._txn:
                cursor.close()

    def get(self, query, *parameters, **kwargs):
        """ Returns the first row returned for the given query."""
        rows = self.query(query, *parameters, **kwargs)
        if not rows:
            return None
        elif len(rows) > 1:
            raise Exception("Multiple rows returned for Database.get() query")
        else:
            return rows[0]

    def execute(self, query, *parameters, **kwargs):
        # If it's an INSERT, return the lastrowid. Otherwise return the rowcount.
        if query.startswith('INSERT'):
            return self.execute_lastrowid(query, *parameters, **kwargs)
        else:
            return self.execute_rowcount(query, *parameters, **kwargs)

    def execute_lastrowid(self, query, *parameters, **kwargs):
        # Executes the given query, returning the lastrowid from the query.
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters)
            if cursor.rowcount > 0:
                try:
                    return cursor.fetchone()[0]
                except PgsqlProgrammingError as ex:
                    logging.error("No Results even though cursor says rowcount > 0. Probably left out '... RETURNING row_id'")
                    return None
        finally:
            if not self._txn:
                cursor.close()

    def execute_rowcount(self, query, *parameters, **kwargs):
        # Executes the given query, returning the rowcount from the query.
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters)
            return cursor.rowcount
        finally:
            if not self._txn:
                cursor.close()

    # Commenting until they can be properly tested for use with commit() and rollback()
    #def executemany(self, query, parameters):
    #    """ Return the lastrowid from the query."""
    #    return self.executemany_lastrowid(query, parameters)
    #
    #def executemany_lastrowid(self, query, parameters):
    #    """ Return the lastrowid from the query."""
    #    cursor = self._cursor()
    #    try:
    #        cursor.executemany(query, parameters)
    #        return cursor.lastrowid
    #    finally:
    #        cursor.close()
    #
    #def executemany_rowcount(self, query, parameters):
    #    """ Return the rowcount from the query."""
    #    cursor = self._cursor()
    #    try:
    #        cursor.executemany(query, parameters)
    #        return cursor.rowcount
    #    finally:
    #        cursor.close()



class MysqlConnection(object):
    """ From torndb originally """
    def __init__(self, host, database, user=None, password=None, max_idle_time=7 * 3600):
        self.host = host
        self.database = database
        self.max_idle_time = max_idle_time
        args = dict(use_unicode=True, charset="utf8mb4", db=database, sql_mode="TRADITIONAL")
        if user is not None:
            args["user"] = user
        if password is not None:
            args["passwd"] = password
        # We accept a path to a MySQL socket file or a host(:port) string
        if "/" in host:
            args["unix_socket"] = host
        else:
            self.socket = None
            pair = host.split(":")
            if len(pair) == 2:
                args["host"] = pair[0]
                args["port"] = int(pair[1])
            else:
                args["host"] = host
                args["port"] = 3306
        self._db_init_command = 'SET time_zone = "+0:00"'
        self._db = None
        self._db_args = args
        self._last_use_time = time.time()
        try:
            self.reconnect()
        except Exception:
            logging.error("Cannot connect to MySQL on %s", self.host, exc_info=True)

    threads = {}

    @classmethod
    def connect(cls, **kwargs):
        if 'mysql_schema' in kwargs:
            host = kwargs['mysql_host']
            schema = kwargs['mysql_schema']
            user = kwargs['mysql_user']
            passwd = kwargs['mysql_password']
        else:
            host = options.mysql_host
            schema = options.mysql_schema
            user = options.mysql_user
            passwd = options.mysql_password
        # force a new connection
        if kwargs.get('force', False):
            return cls(host, schema, user, passwd)
        # check existing connections
        ident = threading.current_thread().ident
        target = '%s@%s' % (schema, host)
        connections = cls.threads.setdefault(ident, {})
        if not target in connections:
            conn = cls(host, schema, user, passwd)
            conn.execute("SET NAMES utf8mb4 COLLATE utf8mb4_unicode_ci", disable_audit_sql=True)
            conn.schema = schema
            cls.threads[ident][target] = conn
        #conn_debug = "%s %s %s" % (ident, target, connections)
        #aegis.stdlib.logw(conn_debug, "CONN DEBUG")
        return connections[target]

    def __del__(self):
        self.close()

    def close(self):
        """Closes this database connection."""
        if getattr(self, "_db", None) is not None:
            self._db.close()
            self._db = None

    def reconnect(self):
        """Closes the existing database connection and re-opens it."""
        self.close()
        self._db = MySQLdb.connect(autocommit=True, connect_timeout=3, **self._db_args)
        self.execute(self._db_init_command, disable_audit_sql=True)

    def iter(self, query, *parameters, **kwargs):
        """Returns an iterator for the given query and parameters."""
        self._ensure_connected()
        cursor = self._db.cursor(MySQLdb.cursors.SSCursor)
        try:
            self._execute(cursor, query, parameters)
            column_names = [d[0] for d in cursor.description]
            if kwargs.get('cls'):
                for row in cursor:
                    yield kwargs['cls'](zip(column_names, row))
            else:
                for row in cursor:
                    yield Row(zip(column_names, row))
        finally:
            cursor.close()

    def query(self, query, *parameters, **kwargs):
        """Returns a row list for the given query and parameters."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters)
            column_names = [d[0] for d in cursor.description]
            if kwargs.get('cls'):
                rows = [kwargs['cls'](zip(column_names, row)) for row in cursor]
            else:
                rows = [Row(zip(column_names, row)) for row in cursor]
            if kwargs.get('return_column_names'):
                return (rows, column_names)
            return rows
        finally:
            cursor.close()

    def get(self, query, *parameters, **kwargs):
        """Returns the first row returned for the given query."""
        rows = self.query(query, *parameters)
        if not rows:
            return None
        elif len(rows) > 1:
            raise Exception("Multiple rows returned for Database.get() query")
        else:
            row = rows[0]
            if row and kwargs.get('cls'):
                row = kwargs['cls'](row)
            return row

    # rowcount is a more reasonable default return value than lastrowid,
    # but for historical compatibility execute() must return lastrowid.
    def execute(self, query, *parameters, **kwargs):
        """Executes the given query, returning the lastrowid from the query."""
        if query.startswith('INSERT'):
            return self.execute_lastrowid(query, *parameters, **kwargs)
        else:
            return self.execute_rowcount(query, *parameters, **kwargs)

    def execute_lastrowid(self, query, *parameters, **kwargs):
        """Executes the given query, returning the lastrowid from the query."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters)
            return cursor.lastrowid
        finally:
            cursor.close()

    def execute_rowcount(self, query, *parameters, **kwargs):
        """Executes the given query, returning the rowcount from the query."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters)
            return cursor.rowcount
        finally:
            cursor.close()

    def executemany(self, query, parameters):
        """Executes the given query against all the given param sequences.
        We return the lastrowid from the query.
        """
        return self.executemany_lastrowid(query, parameters)

    def executemany_lastrowid(self, query, parameters):
        """Executes the given query against all the given param sequences.
        We return the lastrowid from the query.
        """
        cursor = self._cursor()
        try:
            cursor.executemany(query, parameters)
            return cursor.lastrowid
        finally:
            cursor.close()

    def executemany_rowcount(self, query, parameters):
        """Executes the given query against all the given param sequences.
        We return the rowcount from the query.
        """
        cursor = self._cursor()
        try:
            cursor.executemany(query, parameters)
            return cursor.rowcount
        finally:
            cursor.close()

    def _ensure_connected(self):
        # Mysql by default closes client connections that are idle for
        # 8 hours, but the client library does not report this fact until
        # you try to perform a query and it fails.  Protect against this
        # case by preemptively closing and reopening the connection
        # if it has been idle for too long (7 hours by default).
        if (self._db is None or (time.time() - self._last_use_time > self.max_idle_time)):
            self._last_use_time = time.time()
            self.reconnect()

    def _cursor(self):
        self._ensure_connected()
        return self._db.cursor()

    def _execute(self, cursor, query, parameters):
        try:
            aegis.stdlib.incr_start(aegis.stdlib.get_timer(), 'database')
            result = cursor.execute(query, parameters)
            aegis.stdlib.incr_stop(aegis.stdlib.get_timer(), 'database')
            return result
        except (MysqlOperationalError, MysqlInterfaceError) as ex:
            logging.error("Error with MySQL on %s. Close connection and raise.", self.host)
            logging.exception(ex)
            self.close()
            raise


# To support inserting something literally, like NOW(), into mini-ORM below
class Literal(str):
    pass


class Row(dict):
    """ A dict that allows for object-like property access syntax."""
    def __getattr__(self, name):
        if name not in self:
            raise AttributeError(name)
        return self[name]

    @classmethod
    def _table_name(cls):
        # Choose table name based on model table_names attribute
        #aegis.stdlib.logw(cls.table_name, "TABLE NAME")
        if hasattr(cls, 'table_names'):
            if aegis.config.get('pg_database'):
                cls.table_name = cls.table_names['pgsql']
            elif aegis.config.get('mysql_database'):
                cls.table_name = cls.table_names['mysql']
        return cls.table_name

    @classmethod
    def logw(cls, msg, value, row_id=''):
        logging.warning("%s: %s %s", msg, value, row_id)

    @classmethod
    def scan_id(cls, column, row_id):
        sql = 'SELECT * FROM %s WHERE %s=%%s' % (cls._table_name(), column)
        return db().query(sql, row_id, cls=cls)

    @classmethod
    def map_items(cls, items, key):
        return cls([(item[key], item) for item in items])

    @classmethod
    def map_id(cls, row_id, where_col, key_col, debug=False):
        items = cls.map_items(cls.scan_id(where_col, row_id), key_col)
        if debug:
            cls.logw("WHERE", where_col, row_id)
            logging.warning("")
            cls.logw("SCAN", cls.scan_id(where_col, row_id), row_id)
            logging.warning("")
            cls.logw("ITEMS", items, row_id)
        return items

    @classmethod
    def get_id(cls, column_id_val, member_id=None, dbconn=None):
        if not column_id_val:
            return None
        sql = 'SELECT * FROM %s WHERE %s=%%s'
        args = [int(column_id_val)]
        if member_id:
            sql = sql + ' AND member_id=%%s'
            args.append(int(member_id))
        sql = sql % (cls._table_name(), cls.id_column)
        if not dbconn:
            dbconn = db()
        val = dbconn.get(sql, *args, cls=cls)
        return val

    @classmethod
    def scan(cls, dbconn=None):
        if not dbconn:
            dbconn = db()
        sql = 'SELECT * FROM %s' % cls.table_name
        return dbconn.query(sql, cls=cls)

    @classmethod
    def scan_ids(cls, row_ids, dbconn=None):
        if not row_ids:
            return []
        if not dbconn:
            dbconn = db()
        # Fail fast if any of the data are bad, to not return confusing results
        try:
            args, fmt = aegis.database.sql_in_format(row_ids, int)
        except ValueError as ex:
            logging.exception(ex)
            logging.error("Bad arguments passed from %s to aegis.database.Row.scan_ids(): %s", aegis.stdlib.get_caller(), row_ids)
            return []
        sql = "SELECT * FROM "+cls._table_name()+" WHERE "+cls.id_column+" IN ("+fmt+")"
        return dbconn.query(sql, *args, cls=cls)

    # kva_split(), insert(), update() together are a mini-ORM in processing arbitrary column-value combinations on a row.
    # define table_name and data_columns to know which are allowed to be set along with user action
    # columns and where are simple dictionaries: {'full_name': "FULL NAME", 'email': 'email@example.com'}
    @staticmethod
    def kva_split(columns):
        keys = []
        values = []
        args = []
        for key, value in columns.items():
            keys.append('%s' % key)
            if isinstance(value, Literal):
                values.append(value)
            else:
                values.append('%s')
                args.append(value)
        return keys, values, args

    @classmethod
    def insert_columns(cls, sql_txt='INSERT INTO %(db_table)s (%(keys)s) VALUES (%(values)s)', dbconn=None, **columns):
        if not dbconn:
            dbconn = db()
        db_table = cls._table_name()
        # Filter out anything that's not in optional, pre-specified list of data columns
        data_columns = hasattr(cls, 'data_columns') and cls.data_columns
        if data_columns:
            columns = dict( [ (key, val) for key, val in columns.items() if key in data_columns] )
        keys, values, args = cls.kva_split(columns)
        if type(dbconn) is PostgresConnection:
            sql_txt += " RETURNING " + cls.id_column
        sql = sql_txt % {'db_table': db_table, 'keys': ', '.join(keys), 'values': ', '.join(values)}
        #aegis.stdlib.logw(sql, "SQL")
        #aegis.stdlib.logw(args, "ARGS")
        return dbconn.execute(sql, *args)

    @classmethod
    def update_columns(cls, columns, where, dbconn=None):
        if not dbconn:
            dbconn = db()
        if not columns:
            logging.debug('Nothing to update. Skipping query')
            return
        db_table = cls._table_name()
        # Filter out anything that's not in optional, pre-specified list of data columns
        data_columns = hasattr(cls, 'data_columns') and cls.data_columns
        if data_columns:
            columns = dict( [ (key,val) for key, val in columns.items() if key in data_columns] )
        # SET clause
        keys, values, args = cls.kva_split(columns)
        set_clause = ', '.join(['%s=%s' % (key, value) for key, value in zip(keys, values)])
        # WHERE clause
        keys, values, args2 = cls.kva_split(where)
        args += args2
        where_clause = ' AND '.join(['%s=%s' % (key, value) for key, value in zip(keys, values)])
        # SQL statement
        sql = 'UPDATE %s SET %s WHERE %s' % (db_table, set_clause, where_clause)
        #aegis.stdlib.logw(sql, "SQL")
        #aegis.stdlib.logw(args, "ARGS")
        return dbconn.execute_rowcount(sql, *args)
